package net.darkmist.defcon18;

import java.io.EOFException;
import java.io.IOException;
import java.io.FilterInputStream;
import java.io.InputStream;

public class MinBlockingInputStream extends FilterInputStream
{
	public MinBlockingInputStream(InputStream in_arg)
	{
		super(in_arg);
	}

	@Override
	public int read(byte[] b) throws IOException
	{
		return read(b,0,b.length);
	}

	@Override
	public int read(byte[] bytes, int off, int len) throws IOException
	{
		int bytesToRead;
		int bytesRead;
		int b;

		if((bytesToRead = this.available())<=0)
		{
			if((b = this.read())<0)	// BLOCKING
				return b;	// eof
			bytes[off] = (byte)b;
			if((bytesToRead = this.available())<=0)
				return 1;
			bytesToRead = Math.min(bytesToRead, len-1);
			if((bytesRead = super.read(bytes, off+1,bytesToRead))<0)
				return 1;	// we got EOF, but we still have one byte
			return bytesRead + 1;
		}
		bytesToRead = Math.min(bytesToRead, len);
		if((bytesRead = super.read(bytes, off, bytesToRead))<0)
			return bytesRead;	// EOF
		return bytesRead;
	}

	private static final byte[] truncate(byte[] orig, int len)
	{
		byte[] ret;

		if(len == 0)
			return Util.EMPTY_BYTE_ARRAY;
		if(len == orig.length)
			return orig;
		ret = new byte[len];
		System.arraycopy(orig, 0, ret, 0, len);
		return ret;
	}

	/**
	 * Reads as many bytes are available without blocking.
	 * @param off The offset into the array returned where 
	 *	data should be written. Allows a header to be
	 *	added later without copying the array around.
	 * @ret null if nothing could be read. byte array
	 *	containing empty off bytes followed by the data
	 *	bytes read.
	 */
	public byte[] readAvailable(int off) throws IOException
	{
		int bytesToRead;
		int bytesRead;
		byte[] bytes;

		if((bytesToRead = this.available())==0)
			return null;
		bytes = new byte[off + bytesToRead];
		if((bytesRead = this.read(bytes, off, bytesToRead))<0)
			throw new EOFException(); // EOF, go figure
		if(bytesRead == bytesToRead)
			return bytes;
		return truncate(bytes, off+bytesRead);
	}

	/**
	 * Reads as many bytes are available without blocking.
	 * Calls readAvailable(0).
	 * @ret null if nothing could be read. byte array
	 *	containing empty off bytes followed by the data
	 *	bytes read.
	 */
	public byte[] readAvailable() throws IOException
	{
		return readAvailable(0);
	}

	/**
	 * Reads at least one byte blocking if necessary.
	 * @param off The offset into the array returned where 
	 *	data should be written. Allows a header to be
	 *	added later without copying the array around.
	 * @ret null on EOF. byte array containing empty off bytes
	 *	followed by the data bytes read.
	 */
	public byte[] readBlockAvailable(int off) throws IOException
	{
		int bytesToRead;
		int bytesRead;
		byte[] bytes;
		int b;

		// first see if anything is available
		if((bytesToRead = this.available())>0)
		{	// bytes available
			// make sure we aren't trying to allocate 
			// a negative array bounds by wrapping
			bytesToRead = Math.min(bytesToRead, Integer.MAX_VALUE-off);
			bytes = new byte[bytesToRead + off];

			// try to read the bytes
			if((bytesRead = super.read(bytes, off, bytesToRead))<0)
				return null;	// EOF
			if(bytesRead == bytesToRead)
				return bytes;	// exact read
			if(bytesRead > 0)
			{	// short read, truncate...
				return truncate(bytes, bytesRead + off);
			}
			// zero read when we had available...
			// fall through to one byte blocking read
		}

		// read one byte blocking
		if((b = super.read())<0)	// BLOCKING
			return null;	// eof

		// check for more...
		if((bytesToRead = this.available())<=0)
		{	// only got one, return it
			bytes = new byte[off+1];
			bytes[off] = (byte)b;
			return bytes;
		}

		// compute length and allocate buffer for bytes
		bytesToRead = Math.min(bytesToRead, Integer.MAX_VALUE-off-1);
		bytes = new byte[off + 1 + bytesToRead];

		// read in the rest skipping the offset and the
		// one byte we already read...
		if((bytesRead = super.read(bytes, off+1, bytesToRead))<=0)
		{	// EOF or zero read
			bytes = new byte[off+1];
			bytes[off] = (byte)b;
			return bytes;
		}
		// include that first byte...
		bytes[off] = (byte)b;
		if(bytesRead == bytesToRead)
			return bytes;	// exact read
		// runcate as necessary
		return truncate(bytes, off+1+bytesRead);
	}

	/**
	 * Reads at least one byte blocking if necessary. Calls
	 * readBlockAvailable(0).
	 * @ret null on EOF. byte array containing empty off bytes
	 *	followed by the data bytes read.
	 */
	public byte[] readBlockAvailable() throws IOException
	{
		return readBlockAvailable(0);
	}
}
